/*
 *                            COPYRIGHT
 *
 *  cschem - modular/flexible schematics editor - sch-rnd (executable)
 *  Copyright (C) 2022 Tibor 'Igor2' Palinkas
 *
 *  (Supported by NLnet NGI0 PET Fund in 2022)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/cschem
 *    lead developer: email to cschem (at) igor2.repo.hu
 *    mailing list: cschem (at) list.repo.hu (send "subscribe")
 */

/* Font selection */

#include <libcschem/config.h>
#include <stdlib.h>

#include <libcschem/cnc_pen.h>
#include <libcschem/cnc_text.h>
#include <libcschem/rotate.h>

#include <genvector/gds_char.h>
#include <genvector/vts0.h>
#include <librnd/core/safe_fs.h>
#include <librnd/core/hidlib_conf.h>
#include <librnd/core/conf.h>
#include <librnd/core/hid.h>
#include <librnd/core/compat_misc.h>
#include <librnd/core/pixmap.h>
#include <librnd/src_3rd/genregex/regex_sei.h>
#include <ttf2bbox/ttf_map.h>
#include <ttf2bbox/ttf2bbox.h>

#include "conf_core.h"
#include "draw.h"

#include "font.h"

static void *ttf_map_cookie;
static htsp_t *ttf_map;
static ttf2bbox_t ttf;

long sch_rnd_font_score_debug;

int sch_rnd_font_resolve_pen(csch_cpen_t *pen)
{
	static int epic_fail = 0;
	ttf_map_font_t *fm;
	htsp_entry_t *e;
	const char *family, *style;

	if (epic_fail)
		return -1;

	if (pen->font_handle_valid)
		return pen->font_handle;

	retry:;
	fm = NULL;

	/* get pen's preferred font */
	if ((pen->font_family != NULL) || (pen->font_style != NULL)) {
		fm = sch_rnd_font_lookup(pen->font_family, pen->font_style);
		if (fm != NULL) {
			family = pen->font_family;
			style = pen->font_style;
		}
	}

	/* fall back to configured default */
	if (fm == NULL) {
		family = conf_core.editor.default_font.family;
		style = conf_core.editor.default_font.style;
		if ((family != NULL) || (style != NULL))
			fm = sch_rnd_font_lookup(family, style);
		if (fm == NULL)
			family = style = NULL;
	}

	/* final fallback: any font we know of */
	if (fm == NULL) {
		printf("FONT: final fallback");
		e = htsp_first(ttf_map);
		if (e != NULL)
			fm = e->value;
		family = "<final fallback>";
		style = "";
	}

	if (fm == NULL) {
		static int warned = 0;
		if (!warned) {
			rnd_message(RND_MSG_ERROR, "Failed to find absolutely _any_ font; text rendering is impossible.\nPlease revise rc/font_dirs in your conf!\n");
			epic_fail = 1;
		}
		return -1;
	}

	if (!fm->handle_valid) {
		int h = -1;

		if (fm->is_font)
			h = ttf2bbox_font_load(&ttf, fm->path);

		if (h < 0) { /* couldn't load fm->path or it was not even a font file */
			rnd_message(RND_MSG_ERROR, "Failed to load preferred font '%s'; trying to load another instead\n", fm->path);
			htsp_pop(ttf_map, fm->path); /* this leaks, but we shouldn't ever fail to load a font anyway */
			goto retry;
		}
		rnd_message(RND_MSG_DEBUG, "Using font file '%s' for %s %s\n", fm->path, family, style);
		fm->handle = h;
		fm->handle_valid = 1;
	}

	pen->font_handle = fm->handle;
	pen->font_handle_valid = 1;
	return pen->font_handle;
}

/* rot is in rad */
static void font_render_pxm(rnd_pixmap_t *dst, ttf2bbox_pixmap_t *tpm, double rot, double cs, double sn, const rnd_color_t *clr)
{
	long n, end = dst->sx * dst->sy, xo, yo, icx, icy, ocx, ocy;
	unsigned char *o, *i;

	dst->tr = dst->tg = 127;
	dst->tb = 128;
	dst->has_transp = 1;

	/* center points */
	icx = tpm->sx/2;
	icy = tpm->sy/2;
	ocx = dst->sx/2;
	ocy = dst->sy/2;

	for(n = 0, o = dst->p, xo = yo = 0; n < end; n++, o+=3) {
		int oor = 0;
		long ixi, iyi;
		double XO = xo - ocx, YO = yo - ocy, XI, YI;

		/* rotate */
		XI = icx + (cs * XO - sn * YO);
		YI = icy + (sn * XO + cs * YO);

		/* get final input pixel address, clamp */
		ixi = rnd_round(XI); iyi = rnd_round(tpm->sy - YI);
		if ((ixi < 0) || (iyi < 0) || (ixi >= tpm->sx) || (iyi >= tpm->sy))
			oor = 1;
		else
			i = tpm->p + (ixi + iyi * tpm->sx);

		if (oor || (i[0] == 0)) {
			/* out pixel is transparent if out-of-range or input pixel is transparent */
			o[0] = dst->tr;
			o[1] = dst->tg;
			o[2] = dst->tb;
		}
		else {
			/* fake alpha using assumed background */
			o[0] = clr->fr * i[0] + rnd_conf.appearance.color.background.fr * (255-i[0]);
			o[1] = clr->fg * i[0] + rnd_conf.appearance.color.background.fg * (255-i[0]);
			o[2] = clr->fb * i[0] + rnd_conf.appearance.color.background.fb * (255-i[0]);
		}

		xo++;
		if (xo >= dst->sx) {
			xo = 0;
			yo++;
		}

#if 0
		if (n%4 == 0) {
			o[0] = 100;
			if (oor)
				o[1] = 200;
		}
#endif

#if 0
		if ((n % dst->sx) == 0) printf("\n");
		printf("%d", i[0]/64);
#endif
	}
}

void sch_rnd_font_text_update(csch_sheet_t *sheet, csch_text_t *t, csch_cpen_t *pen)
{
	int font_handle;

	if (t->pixmap == NULL) {
		ttf2bbox_pixmap_t *tpm;
		double rot;
		double cs, sn;
		const char *text_str;

		font_handle = sch_rnd_font_resolve_pen(pen);
		if (font_handle < 0)
			return;

		text_str = csch_text_get_rtext(t);

	/* figure which kind of text fitting to use */
		if (t->has_bbox) {
			t->crd_sx = t->hdr.bbox.x2 - t->hdr.bbox.x1;
			t->crd_sy = t->hdr.bbox.y2 - t->hdr.bbox.y1;
			t->pix_sx = RND_ABS(t->spec2.x - t->spec1.x) * conf_core.appearance.font_resolution;
			t->pix_sy = RND_ABS(t->spec2.y - t->spec1.y) * conf_core.appearance.font_resolution;
		}
		else {
			long tox, toy, sx, sy;
			double scale = (double)pen->font_height * conf_core.appearance.font_resolution;
/*rnd_trace("pen height=%ld; est height: %f handle=%d\n", (long)pen->font_height, scale, font_handle);*/
			if (ttf2bbox_estimate_at(&ttf, font_handle, scale, text_str, 0, 0, 0, TTF2BBOX_LEFT, TTF2BBOX_BOTTOM, &tox, &toy, &sx, &sy, 0) < 0) {
				rnd_message(RND_MSG_ERROR, "Failed to estimate text size with ttf2bbox\n"); /* should never happen */
				return;
			}
			t->pix_sx = sx;
			t->pix_sy = sy;
			t->spec2.x = t->spec1.x + t->pix_sx / conf_core.appearance.font_resolution;
			t->spec2.y = t->spec1.y + pen->font_height;
			csch_text_update(sheet, t, 1);
			t->crd_sx = t->hdr.bbox.x2 - t->hdr.bbox.x1;
			t->crd_sy = t->hdr.bbox.y2 - t->hdr.bbox.y1;

/*rnd_trace("pix sx=%ld sy=%ld crd sx=%ld sy=%ld\n", sx, sy, t->crd_sx, t->crd_sy);*/

		}

		rot = t->inst_rot / RND_RAD_TO_DEG;
		if (t->hdr.parent->xform.mirx ^ t->hdr.parent->xform.miry)
			rot = -rot;

		cs = cos(rot); sn = sin(rot);


		tpm = ttf2bbox_pm_alloc(t->pix_sx, t->pix_sy);
		ttf2bbox_render_in_box(&ttf, font_handle, tpm, text_str, 0, 0, 0, t->pix_sx, t->pix_sy, 0);

/*rnd_trace("final pixmap1: %ld %ld\n", t->pix_sx, t->pix_sy);*/
		{ /* calculate a bbox that hosts all 4 corners of the rotated text box */
			double sx1, sy1, sx2, sy2, x = t->pix_sx/2.0, y = t->pix_sy/2.0;

			sx1 = fabs(cs * x - sn * y);
			sy1 = fabs(sn * x + cs * y);
			sx2 = fabs(cs * x + sn * y);
			sy2 = fabs(sn * x - cs * y);
			t->pix_sx = (sx1 > sx2 ? sx1 : sx2) * 2.0;
			t->pix_sy = (sy1 > sy2 ? sy1 : sy2) * 2.0;
		}

		t->pixmap = rnd_pixmap_alloc(&sheet->hidlib, t->pix_sx, t->pix_sy);
/*rnd_trace("final pixmap2: %ld %ld\n", t->pix_sx, t->pix_sy);*/
		font_render_pxm(t->pixmap, tpm, rot - G2D_PI, cs, -sn, (csch_chdr_is_selected(&t->hdr) ? &conf_core.appearance.color.selected_stroke : &pen->color));
		free(tpm);
	}
}

void sch_rnd_font_text_render(csch_sheet_t *sheet, rnd_hid_gc_t gc, csch_text_t *t, csch_cpen_t *pen)
{
	csch_coord_t cx, cy;

	if ((rnd_render->draw_pixmap == NULL) || (t->pixmap == NULL))
		return;

	cx = (t->hdr.bbox.x1 + t->hdr.bbox.x2)/2;
	cy = (t->hdr.bbox.y1 + t->hdr.bbox.y2)/2;

	if ((t->crd_sx > 0) && (t->crd_sy > 0))
		rnd_render->draw_pixmap(rnd_render, C2P(cx), C2P(cy), C2P(t->crd_sx), C2P(t->crd_sy), t->pixmap);
}

static void text_inval_font(csch_sheet_t *sheet, csch_text_t *text)
{
	if (text->pixmap == NULL)
		return;

	if (rnd_render->uninit_pixmap != NULL)
		rnd_render->uninit_pixmap(rnd_render, text->pixmap);
	
	rnd_pixmap_free(text->pixmap);
	text->pixmap = NULL;
}

static void text_calc_bbox(csch_sheet_t *sheet, csch_text_t *text)
{
	csch_cpen_t *pen = csch_stroke_fallback_(&text->hdr, NULL);
	sch_rnd_font_text_update(sheet, text, pen);
}

static void pen_inval_font(csch_sheet_t *sheet, csch_cpen_t *pen)
{
	pen->font_handle_valid = 0;
}


static struct ttf_map_font_s *sch_rnd_font_lookup_(const char *name, const char *style)
{
	ttf_map_font_t *best = NULL;
	long best_score = -1000000l;
	long namelen, stylelen;
	htsp_entry_t *e;
	re_sei_t *rxname, *rxstyle;

	if (name == NULL)
		name = ".";
	if (style == NULL)
		style = ".";

	rxname = re_sei_comp(name);
	rxstyle = re_sei_comp(style);
	namelen = strlen(name);
	stylelen = strlen(style);

	for(e = htsp_first(ttf_map); e != NULL; e = htsp_next(ttf_map, e)) {
		ttf_map_font_t *f = e->value;
		long score = 0;

		if (!f->is_font)
			continue;

		if (rxname != NULL) {
			long fambad = strlen(f->family_name) - namelen;
			long pathbad = strlen(f->path) - namelen;
			score += re_sei_exec(rxname, f->family_name) * 8 - fambad*fambad * 8;
			score += re_sei_exec(rxname, f->path) - pathbad*pathbad/2;
		}
		if (rxstyle != NULL) {
			long stylebad = strlen(f->style_name) - stylelen;
			score += re_sei_exec(rxstyle, f->style_name) * 8 - stylebad*stylebad*16;
		}
		

		if (sch_rnd_font_score_debug)
			fprintf(stderr, "%ld FONT: '%s' ('%s' '%s')\n", score, f->path, f->family_name, f->style_name);
		if (score > best_score) {
			best_score = score;
			best = f;
		}
	}

	if (rxname != NULL)
		re_sei_free(rxname);
	if (rxstyle != NULL)
		re_sei_free(rxstyle);

	return best;
}

struct ttf_map_font_s *sch_rnd_font_lookup(const char *name, const char *style)
{
	struct ttf_map_font_s *res = sch_rnd_font_lookup_(name, style);

	if (res == NULL) /* fall back to configured default */
		res = sch_rnd_font_lookup_(conf_core.editor.default_font.family, conf_core.editor.default_font.style);

	if (res == NULL) { /* fall back to anything */
		htsp_entry_t *e = htsp_first(ttf_map);
		if (e != NULL)
			res = e->value;
	}
	return res;
}

void sch_rnd_font_init2(void)
{
	rnd_conf_listitem_t *ci;
	gds_t tmp = {0};
	vts0_t fdirs = {0};

	gds_append_str(&tmp, rnd_conf.rc.path.home);
	gds_append(&tmp, '/');
	gds_append_str(&tmp, DOT_SCH_RND);

	if (!rnd_is_dir(NULL, tmp.array))
		rnd_mkdir(NULL, tmp.array, 0755);

	gds_append_str(&tmp, "/ttf_map.cache");
	for(ci = rnd_conflist_first((rnd_conflist_t *)&conf_core.rc.font_dirs); ci != NULL; ci = rnd_conflist_next(ci)) {
		const char *dr = ci->val.string[0];
		vts0_append(&fdirs, (char *)dr);
	}

	if (fdirs.array == NULL) {
		fprintf(stderr, "No font dir configured. Please set rc/font_dirs to point to at least one directory hosting fonts\n");
		exit(1);
	}

	ttf_map = ttf_map_load(tmp.array, (const char **)fdirs.array, &ttf_map_cookie);

	gds_uninit(&tmp);
	vts0_uninit(&fdirs);


	if (ttf2bbox_init(&ttf) != 0) {
		fprintf(stderr, "Failed to initialize ttf2bbox\n"); /* shuld never happen */
		exit(1);
	}
	ttf2bbox_y_mirror = 0;
	csch_cb_text_calc_bbox = text_calc_bbox;
	csch_cb_text_invalidate_font = text_inval_font;
	csch_cb_pen_invalidate_font = pen_inval_font;
}

void sch_rnd_font_uninit(void)
{
	ttf_map_free(ttf_map, &ttf_map_cookie);
	ttf2bbox_uninit(&ttf);
	csch_cb_text_calc_bbox = NULL;
	csch_cb_text_invalidate_font = NULL;
	csch_cb_pen_invalidate_font = NULL;
}
